import time

import numpy as np

from dwave.system.samplers import DWaveSampler

# Numpy arrays should be row-major for best performance


class TrainEnv:
    # require for the data to be passed as NP data arrays
    # -> clear to both user and code what each array is
    # X_train, y_train should be formatted like scikit-learn data arrays are
    # -> X_train is the train data, y_train is the train labels
    def __init__(
        self,
        X_train,
        y_train,
        endpoint_url,
        account_token,
        X_val=None,
        y_val=None,
        fidelity=7,
    ):
        self.X_train = X_train
        self.y_train = y_train
        if X_val is None:
            self.create_val_data()
        else:
            self.X_val = X_val
            self.y_val = y_val
        self.train_size = np.shape(self.X_train)[0]

        self.fidelity = fidelity
        self.fidelity_offset = 0.0225 / fidelity

        self.c_i = None
        self.C_i = None
        self.C_ij = None
        self.data_preprocess()

        self.sampler = DWaveSampler(
            endpoint=endpoint_url,
            token=account_token,
            solver=dict(topology__type="pegasus"),
        )

    def create_val_data(self):
        dummy_xt, dummy_xv = np.split(
            self.X_train, [int(8 * np.size(self.X_train, 0) / 10)], 0
        )
        dummy_yt, dummy_yv = np.split(
            self.y_train, [int(8 * np.size(self.y_train) / 10)]
        )

        self.X_train, self.X_val = np.array(list(dummy_xt)), np.array(list(dummy_xv))
        self.y_train, self.y_val = np.array(list(dummy_yt)), np.array(list(dummy_yv))

    def data_preprocess(self):
        """
        This duplicates the parameters 'fidelity' times. The purpose is to turn the weak classifiers
        from outputing a single number (-1 or 1) to outputting a binary array ([-1, 1, 1,...]). The
        use of such a change is to trick the math into allowing more nuance between a weak classifier
        that outputs 0.1 from a weak classifier that outputs 0.9 (the weak classifier outputs are continuous)
        -> thereby discretizing the weak classifier's decision into more pieces than binary.
        """
        """
            This creates a periodic array to shift the outputs of the repeated weak classifier, so that there
            is a meaning to duplicating them. You can think of each successive digit of the resulting weak classifier
            output array as being more specific about what the continuous output was - ie >0, >0.1, >0.2 etc. This 
            description is not exactly correct in this case but it is the same idea as what we're doing.
        """
        m_events, n_params = np.shape(
            self.X_train
        )  # [M events (rows) x N parameters (columns)]

        c_i = np.repeat(
            self.X_train, repeats=self.fidelity, axis=1
        )  # [M events (rows) x N*fidelity parameters (columns)]

        offset_array = self.fidelity_offset * (
            np.tile(np.arange(self.fidelity), m_events * n_params) - self.fidelity // 2
        )
        c_i = np.sign(np.ndarray.flatten(c_i, order="C") - offset_array) / (
            n_params * self.fidelity
        )
        c_i = np.reshape(c_i, (m_events, n_params * self.fidelity))
        self.c_i = c_i

        C_i = np.einsum("i, ij", self.y_train, c_i)
        C_ij = np.einsum("ji,jk", c_i, c_i)

        self.C_i, self.C_ij = C_i, C_ij



import datetime

import numpy as np

from sklearn.metrics import accuracy_score

from .anneal import anneal
from .anneal import default_prune, default_qac, decode_qac, energies_qac, uniques_qac

# Used to calculate the total hamiltonian of a certain problem
def total_hamiltonian(mu, s, sigma, C_i, C_ij):
    """Derived from Eq. 9 in QAML-Z paper (ZLokapa et al.)
    Dot products of upper triangle

    TODO: Check indecies; maybe add .T
    """
    ham = np.einsum("i, i", -C_i, s * sigma)
    ham = ham + np.einsum("i, i", np.einsum("ij, j", np.triu(C_ij, k=1), mu), s * sigma)
    ham = ham + np.einsum(
        "i, i", np.einsum("i, ij", s * sigma, np.triu(C_ij, k=1)), s * sigma
    )

    return ham


# Returns the ML algorithm's predictions
def evaluate(X_data, weights):
    """
    TODO:
    """
    return np.sign(np.dot(weights, X_data))


class ModelConfig:
    def __init__(self, n_iterations=10, zoom_factor=0.5):
        self.n_iterations = n_iterations
        self.zoom_factor = zoom_factor
        self.anneal_time = 5

        self.flip_higher_probs = np.array(
            [0.16, 0.08, 0.04, 0.02] + [0.01] * (n_iterations - 4)
        )
        self.flip_lower_probs = self.flip_higher_probs / 2

        self.strengths = [3.0, 1.0, 0.5, 0.2] + [0.1] * (n_iterations - 4)
        self.energy_fractions = [0.08, 0.04, 0.02] + [0.01] * (n_iterations - 3)
        self.ngauges = [50, 10] + [1] * (n_iterations - 2)
        self.max_states = [16, 4] + [1] * (n_iterations - 2)
        self.nread = 200

        self.embedding = None
        self.fixed_dict = None
        self.same_fixed_dict_counter = 0
        self.diff_fixed_dict_counter = 0

        self.fix_vars = True
        self.prune_vars = default_prune
        self.cutoff = 95
        self.encode_vars = default_qac
        self.encoding_depth = 3  # from nested qac paper
        self.decode_vars = decode_qac
        self.encoded_energies = energies_qac
        self.encoded_uniques = uniques_qac


class Model:
    """
    TODO: finish results dict -> add in method for final test accuracy (AUROC) computation
    """

    def __init__(self, config, env):
        # add in hyperparameters in ModelConfig
        # -> this is where user determines how the model will train
        self.config = config
        self.env = env

        self.start_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        self.anneal_results = {}
        self.mus_dict = {}

    def pick_excited_states(self, iter, excited_states, mu):
        for excited_state in excited_states:
            new_sigma = pow(self.config.zoom_factor, iter + 1)
            new_energy = total_hamiltonian(
                mu, excited_state, new_sigma, self.env.C_i, self.env.C_ij
            )
            flips = np.ones(np.size(excited_state))
            for state in range(np.size(excited_state)):
                test_state = np.copy(excited_state)
                test_state[state] = 0
                test_energy = total_hamiltonian(
                    mu, test_state, new_sigma, self.env.C_i, self.env.C_ij
                )
                if new_energy > test_energy:
                    flip_prob = self.config.flip_higher_probs[iter]
                else:
                    flip_prob = self.config.flip_lower_probs[iter]
                flip = np.random.choice([-1, 1], p=[flip_prob, 1 - flip_prob])
                flips[state] = flip
            new_state = excited_state * flips

        return mu + new_state * new_sigma

    def train(self):
        """
        TODO:
        """
        mus = [np.ones(np.size(self.env.C_i))]

        for i in range(self.config.n_iterations):
            new_mus_arr = []
            for mu in mus:
                excited_states_arr = anneal(self.config, i, self.env, mu)
                for excited_states in excited_states_arr:
                    new_mus_arr.append(
                        [self.pick_excited_states(i, excited_states, mu)]
                    )
            accuracies = np.zeros(len(new_mus_arr))
            ind = 0
            for new_mus in new_mus_arr:
                avg_arr_val = []
                for new_mu in new_mus:
                    avg_arr_val.append(
                        accuracy_score(self.env.y_train, evaluate(new_mu, self.env.c_i))
                    )
                accuracies[ind] = np.mean(np.array(avg_arr_val))
                ind += 1
            chosen_new_mus = np.argmax(accuracies)
            self.anneal_results["iter%d" % i] = accuracies[chosen_new_mus]
            mus = new_mus_arr[chosen_new_mus]
            mus_filename = "mus%05d-%d_iter%d_run%d__%s" % (
                self.env.train_size,
                self.env.fidelity,
                i,
                chosen_new_mus,
                self.start_time,
            )
            self.mus_dict[mus_filename] = mus
